import feedparser
from audiobooker import AudioBook, BookGenre, BookAuthor
from audiobooker.scrappers import AudioBookSource
from audiobooker.utils.google_search import GoogleSearch
from threading import Thread


class LoyalBooksAudioBook(AudioBook):
    base_url = "http://www.loyalbooks.com"

    def __init__(self, title="", authors=None, description="", genres=None,
                 book_id="", runtime=0, url="", rss_url="", img="", rating=0,
                 language='english', from_data=None):
        """

        Args:
            title:
            authors:
            description:
            genres:
            book_id:
            runtime:
            url:
            rss_url:
            language:
            from_data:
        """
        AudioBook.__init__(self, title, authors, description, genres,
                           book_id, runtime, url, img, language)
        self.rss_url = rss_url or url + "/feed"
        self.rating = rating
        if not self.book_id and self.url:
            self.book_id = self.url.split("/")[-1]
        if from_data:
            self.from_json(from_data)
        self.raw = from_data or {}
        self.from_rss()
        self.from_page()

    def parse_page(self):
        title = self.soup.find("span", {"itemprop": "name"}).text
        description = self.soup.find("font",
                                     {"class": "book-description"}).text
        if self.soup.find(id="star1") is not None:
            rating = 1
        elif self.soup.find(id="star2") is not None:
            rating = 2
        elif self.soup.find(id="star3") is not None:
            rating = 3
        elif self.soup.find(id="star4") is not None:
            rating = 4
        elif self.soup.find(id="star5") is not None:
            rating = 5
        else:
            rating = 0
        author = self.soup.find("font", {"class": "book-author"})
        author_name = author.text.replace("By: ", "")

        names = author_name.split(" ")
        if len(names):
            first_name = names[0].strip()
            last_name = " ".join(names[1:]).strip()
            if not last_name:
                last_name = first_name
                first_name = ""
        else:
            first_name = ""
            last_name = author_name.strip()

        author_url = author.find("a")
        if author_url:
            author_url = self.base_url + author_url["href"]

        authors = [BookAuthor(url=author_url, first_name=first_name,
                              last_name=last_name)]

        genres = []
        genres_table = self.soup.find(summary="Genres for this book")
        if genres_table:
            genres_urls = genres_table.find_all("a")
            for a in genres_urls:
                url = self.base_url + a["href"]
                genre = a.text.strip()
                genre_id = LoyalBooks.get_genre_id(genre)
                genres.append(BookGenre(name=genre, url=url,
                                        genre_id=genre_id))

        img = self.soup.find("img", {"itemprop": "image", "class": "cover"})
        if img:
            img = self.base_url + img["src"]
        return {"description": description, "rating": rating, "genres": genres,
                "authors": authors, "title": title, "img": img}

    @property
    def rss_data(self):
        """

        Returns:

        """
        return feedparser.parse(self.rss_url)

    @property
    def streamer(self):
        """

        """
        for stream in self.rss_data["entries"]:
            try:
                for url in stream["links"]:
                    if url["type"] == 'audio/mpeg':
                        yield url["href"]
            except Exception as e:
                print(e)
                continue

    @property
    def authors(self):
        """

        Returns:

        """
        return [BookAuthor(from_data=a) for a in self._authors]

    @property
    def genres(self):
        """

        Returns:

        """
        return [BookGenre(from_data=a) for a in self._genres]

    def from_json(self, json_data):
        """

        Args:
            json_data:
        """
        AudioBook.from_json(self, json_data)
        self.rss_url = json_data.get("url_rss", self.rss_url)
        self.rating = json_data.get("rating", self.rating)
        if not self.book_id and self.url:
            self.book_id = self.url.split("/")[-1]

    def calc_runtime(self, data=None):
        data = data or self.rss_data["entries"]
        for rss_data in data:
            runtime = rss_data["itunes_duration"].split(":")
            if len(runtime) == 1:  # seconds
                self.runtime += int(runtime[0])
            elif len(runtime) == 2:  # minutes : seconds
                self.runtime += int(runtime[1]) + (int(runtime[0]) * 60)
            elif len(runtime) == 3:  # hours : minutes : seconds
                self.runtime += int(runtime[2]) + (int(runtime[1]) * 60) + \
                                (int(runtime[0]) * 120)

    def from_rss(self):
        rss = self.rss_data["entries"]

        if self.runtime < 1:
            self.calc_runtime()

        if not self.url:
            self.url = rss[0]["link"]

        for rss_data in rss:
            first_name = ""
            last_name = rss_data["author"]
            names = last_name.split(" ")
            if len(names) > 1:
                first_name = names[0].strip()
                last_name = " ".join(names[1:]).strip()
                if not last_name:
                    last_name = first_name
                    first_name = ""
            author = BookAuthor(from_data={"first_name": first_name,
                                           "last_name": last_name})
            if author.as_json not in self._authors:
                self._authors.append(author.as_json)

    def from_page(self):
        data = self.parse_page()
        if self.rating < 1:
            self.rating = data["rating"]
        if not self.title:
            self.title = data["title"]
        if not self._description:
            self._description = data["description"]

        self.img = data.get("img", self.img)
        for genre in data["genres"]:
            if genre.as_json not in self._genres:
                self._genres.append(genre.as_json)
        for author in data["authors"]:
            if author.as_json not in self._authors:
                self._authors.append(author.as_json)

    @property
    def as_json(self):
        bucket = self.raw
        bucket["url"] = self.url
        bucket["rss_url"] = self.rss_url
        bucket["img"] = self.img
        bucket["title"] = self.title
        bucket["authors"] = self._authors
        bucket["description"] = self._description
        bucket["genres"] = self._genres
        bucket["id"] = self.book_id
        bucket["runtime"] = self.runtime
        bucket["language"] = self.lang
        bucket["rating"] = self.rating
        bucket["streams"] = [s for s in self.streamer]
        return bucket

    def __repr__(self):
        """

        Returns:

        """
        return "LoyalBooksAudioBook(" + str(self) + ", " + self.book_id + ")"


class LoyalBooks(AudioBookSource):
    base_url = "http://www.loyalbooks.com"
    popular_url = "http://www.loyalbooks.com"
    genres_url = "http://www.loyalbooks.com/genre-menu"
    search_url = "http://www.loyalbooks.com/search?q=%s"
    _genres = None
    _genre_pages = None

    @staticmethod
    def scrap_genres():
        soup = LoyalBooks._get_soup(
            LoyalBooks._get_html(LoyalBooks.genres_url))
        urls = soup.find("div", {"class": "left"}).find_all("a")
        bucket = {}
        for url in urls:
            genre = url.text
            url = url["href"]
            if url.startswith("/genre"):
                url = "http://www.loyalbooks.com" + url
                bucket[genre] = url
        return bucket

    @staticmethod
    def get_genre(genre_id):
        """

        Args:
            genre_id:

        Returns:
            BookGenre

        """
        genre = ""
        if LoyalBooks._genres is not None:
            if genre_id <= len(LoyalBooks._genres):
                genre = LoyalBooks._genres[genre_id]
        else:
            genres = []
            for genre in LoyalBooks.scrap_genres():
                genres.append(genre)
            genres = sorted(genres)
            genre = genres[genre_id]
        return BookGenre(genre_id=genre_id, name=genre)

    @staticmethod
    def get_genre_id(genre):
        if LoyalBooks._genres is not None:
            return str(LoyalBooks._genres.index(genre))
        genres = []
        for gen in LoyalBooks.scrap_genres():
            genres.append(gen)
        genres = sorted(genres)
        return str(genres.index(genre))

    @property
    def genre_pages(self):
        if LoyalBooks._genre_pages is None:
            try:
                LoyalBooks._genre_pages = LoyalBooks.scrap_genres()
            except Exception as e:
                LoyalBooks._genre_pages = {
                    'Adventure': 'http://www.loyalbooks.com/genre/Adventure',
                    'Advice': 'http://www.loyalbooks.com/genre/Advice',
                    'Ancient Texts': 'http://www.loyalbooks.com/genre/Ancient_Texts',
                    'Animals': 'http://www.loyalbooks.com/genre/Animals',
                    'Art': 'http://www.loyalbooks.com/genre/Art',
                    'Biography': 'http://www.loyalbooks.com/genre/Biography',
                    'Children': 'http://www.loyalbooks.com/genre/Children',
                    'Classics (antiquity)': 'http://www.loyalbooks.com/genre/Classics_antiquity',
                    'Comedy': 'http://www.loyalbooks.com/genre/Comedy',
                    'Cookery': 'http://www.loyalbooks.com/genre/Cookery',
                    'Dramatic Works': 'http://www.loyalbooks.com/genre/Dramatic_Works',
                    'Economics': 'http://www.loyalbooks.com/genre/Economics_Political_Economy',
                    'Epistolary fiction': 'http://www.loyalbooks.com/genre/Epistolary_fiction',
                    'Essay/Short nonfiction': 'http://www.loyalbooks.com/genre/Essay_Short_nonfiction',
                    'Fairy tales': 'http://www.loyalbooks.com/genre/Fairy_tales',
                    'Fantasy': 'http://www.loyalbooks.com/genre/Fantasy',
                    'Fiction': 'http://www.loyalbooks.com/genre/Fiction',
                    'Historical Fiction': 'http://www.loyalbooks.com/genre/Historical_Fiction',
                    'History': 'http://www.loyalbooks.com/genre/History',
                    'Holiday': 'http://www.loyalbooks.com/genre/Holiday',
                    'Horror/Ghost stories': 'http://www.loyalbooks.com/genre/Horror_Ghost_stories',
                    'Humor': 'http://www.loyalbooks.com/genre/Humor',
                    'Instruction': 'http://www.loyalbooks.com/genre/Instruction',
                    'Languages': 'http://www.loyalbooks.com/genre/Languages',
                    'Literature': 'http://www.loyalbooks.com/genre/Literature',
                    'Memoirs': 'http://www.loyalbooks.com/genre/Memoirs',
                    'Music': 'http://www.loyalbooks.com/genre/Music',
                    'Mystery': 'http://www.loyalbooks.com/genre/Mystery',
                    'Myths/Legends': 'http://www.loyalbooks.com/genre/Myths_Legends',
                    'Nature': 'http://www.loyalbooks.com/genre/Nature',
                    'Non-fiction': 'http://www.loyalbooks.com/genre/Non-fiction',
                    'Philosophy': 'http://www.loyalbooks.com/genre/Philosophy',
                    'Play': 'http://www.loyalbooks.com/genre/Play',
                    'Poetry': 'http://www.loyalbooks.com/genre/Poetry',
                    'Politics': 'http://www.loyalbooks.com/genre/Politics',
                    'Psychology': 'http://www.loyalbooks.com/genre/Psychology',
                    'Religion': 'http://www.loyalbooks.com/genre/Religion',
                    'Romance': 'http://www.loyalbooks.com/genre/Romance',
                    'Satire': 'http://www.loyalbooks.com/genre/Satire',
                    'Science': 'http://www.loyalbooks.com/genre/Science',
                    'Science fiction': 'http://www.loyalbooks.com/genre/Science_fiction',
                    'Sea stories': 'http://www.loyalbooks.com/genre/Sea_stories',
                    'Self Published': 'http://www.loyalbooks.com/genre/Self-Published',
                    'Short stories': 'http://www.loyalbooks.com/genre/Short_stories',
                    'Spy stories': 'http://www.loyalbooks.com/genre/Spy_stories',
                    'Teen/Young adult': 'http://www.loyalbooks.com/genre/Teen_Young_adult',
                    'Tragedy': 'http://www.loyalbooks.com/genre/Tragedy',
                    'Travel': 'http://www.loyalbooks.com/genre/Travel',
                    'War stories': 'http://www.loyalbooks.com/genre/War_stories',
                    'Westerns': 'http://www.loyalbooks.com/genre/Westerns'}
        return self._genre_pages or {}

    @property
    def genres(self):
        if LoyalBooks._genres is None:
            try:
                LoyalBooks._genres = list(LoyalBooks.genre_pages.keys())
            except Exception as e:
                LoyalBooks._genres = ['Advice', 'Instruction', 'Ancient ' \
                                                               'Texts',
                                      'Biography', 'Memoirs', 'Languages',
                                      'Myths/Legends', 'Holiday', 'Art',
                                      'Politics', 'Short stories', 'Romance',
                                      'Essay/Short nonfiction', 'Fiction',
                                      'Epistolary fiction', 'Science',
                                      'Nature', 'Dramatic Works',
                                      'Spy stories', 'History', 'Non-fiction',
                                      'Historical Fiction', 'Play', 'Children',
                                      'Satire', 'Humor',
                                      'Classics (antiquity)', 'Travel',
                                      'Religion', 'Adventure', 'Animals',
                                      'Psychology', 'Sea stories',
                                      'Horror/Ghost stories', 'Fantasy',
                                      'Cookery', 'Poetry', 'Self Published',
                                      'Westerns', 'Comedy', 'Music',
                                      'Economics', 'Fairy tales', 'Tragedy',
                                      'Teen/Young adult', 'Literature',
                                      'War stories', 'Science fiction',
                                      'Philosophy', 'Mystery']
        return sorted(self._genres) or []

    def _parse_book_div(self, book):
        try:
            url = self.base_url + book.find("a")[
                "href"].strip()
            img = book.find("img")
            if img:
                img = self.base_url + img["src"].strip()
            name = book.find("b")
            if name:
                name = name.text.strip()
                author = book.text.replace(name, "").strip()
            else:
                name, author = book.find("div", {"class": "s-left"}) \
                    .text.split(" By: ")
            if book.find(id="star1") is not None:
                rating = 1
            elif book.find(id="star2") is not None:
                rating = 2
            elif book.find(id="star3") is not None:
                rating = 3
            elif book.find(id="star4") is not None:
                rating = 4
            elif book.find(id="star5") is not None:
                rating = 5
            else:
                rating = 0
            names = author.split(" ")
            if len(names):
                first_name = names[0].strip()
                last_name = " ".join(names[1:]).strip()
                if not last_name:
                    last_name = first_name
                    first_name = ""
            else:
                first_name = ""
                last_name = author.strip()
            return LoyalBooksAudioBook(title=name.strip(), url=url,
                                       img=img or "", rating=rating,
                                       authors=[BookAuthor(
                                           first_name=first_name,
                                           last_name=last_name).as_json])
        except Exception as e:
            pass  # probably an add
        return None

    def scrap_by_genre(self, genre, limit=-1, offset=0):
        """

        Generator, yields AudioBook objects

        Args:
            genre:
            limit:
            offset:
        """
        url = self.genre_pages[genre] + "?page=" + str(offset)
        limit = int(limit)
        soup = self._get_soup(self._get_html(url))
        el = soup.find("table", {"class": "layout2-blue"})
        if el is None:
            el = soup.find("table", {"class": "layout3"})

        books = el.find_all("td", {"class": "layout2-blue"})
        if not len(books):
            books = el.find_all("td", {"class": "layout3"})

        for book in books:
            book = self._parse_book_div(book)
            if book is None:
                continue
            book._genres = [BookGenre(name=genre, url=self.genre_pages[genre],
                                      genre_id=self.get_genre_id(
                                          genre)).as_json],
            yield book

        # check if last page reached
        pages = soup.find("div", {"class": "result-pages"}).text
        if ">" not in pages:
            return

        # check if limit crawled
        if limit > 0 and int(offset) > limit:
            return

        # crawl next page
        for book in self.scrap_by_genre(genre, offset + 1, limit):
            yield book

    def scrap_popular(self, limit=-1, offset=0):
        """

        Generator, yields AudioBook objects

        Args:
            limit:
            offset:
        """
        soup = self._get_soup(self._get_html(self.popular_url))
        books = soup.find(summary="Audio books").find_all("td")
        for b in books:
            b = self._parse_book_div(b)
            if b is not None:
                yield b

    @staticmethod
    def search_audiobooks(since=None, author=None, title=None, genre=None,
                          limit=25):
        """

        Args:
            since: a UNIX timestamp; returns all projects cataloged since that time
            author: all records by that author last name
            title: all matching titles
            genre: all projects of the matching genre

        Yields:
            AudioBook objects
        """
        query = ""
        if title:
            query += title + " "

        if genre:
            query += genre + " "

        if author:
            query += author + " "

        ## TODO find out how to get callback and nocache values
        """
        import requests
        url = LoyalBooks.search_url % query
        cx = "003017802411926626169:x3dul6qfjls"
        session = requests.Session()
        session.get(url)
        cx_token = session.get(
            "https://cse.google.com/cse.js?cx=003017802411926626169"
            ":x3dul6qfjls").text.split('"cse_token": "')[1].split('",')[0]
        params = {"rsz": "filtered_cse",
                  "num": "10",
                  "hl": "en",
                  "source": "gcsc",
                  "gss": ".com",
                  "cx": cx,
                  "q": query,
                  "safe": "off",
                  "cse_tok": cx_token,
                  "sort": ""#,
                  #"callback": "google.search.cse.api15358",
                  #"nocache": "1546515546857"
                  }
        print(session.get("https://cse.google.com/cse/element/v1",
                          data=params))
        """

        query += " site:" + LoyalBooks.base_url

        for result in GoogleSearch.search(query, limit):
            if "www.loyalbooks.com/book/" not in result.url:
                continue
            if result.url.endswith("/feed"):
                continue
            yield LoyalBooksAudioBook(url=result.url)

        return []

    @staticmethod
    def get_audiobook(book_id):
        """

        Args:
            book_id:

        Returns:
            AudioBook

        """
        url = 'http://www.loyalbooks.com/book/' + book_id
        book = LoyalBooksAudioBook(url=url)
        return book

    def scrap_all_audiobooks(self, limit=-1, offset=0):
        """

        Generator, yields AudioBook objects

        Args:
            limit:
            offset:
        """
        for genre in self.genres:
            for book in self.scrap_by_genre(genre, limit, offset):
                yield book


if __name__ == "__main__":
    from pprint import pprint

    book = LoyalBooks.get_audiobook('Slave-Is-A-Slave-by-H-Beam-Piper')
    pprint(book.parse_page())

    for a in book.authors:
        print(a.as_json)

    print(LoyalBooks.get_genre(40))

    for book in LoyalBooks.search_audiobooks(author="Lovecraft"):
        pprint(book.as_json)

    scraper = LoyalBooks()
    for book in scraper.scrap_popular():
        pprint(book.as_json)

    for book in scraper.scrap_by_genre("Science fiction"):
        pprint(book.as_json)

    for book in scraper.scrap_all_audiobooks():
        pprint(book.as_json)
    pprint(scraper.scrap_genres())
    pprint(scraper.genres)



import requests
import feedparser
from audiobooker import AudioBook, BookGenre, BookAuthor
from audiobooker.scrappers import AudioBookSource


class LibrivoxAudioBook(AudioBook):
    """
    """

    def __init__(self, title="", authors=None, description="", genres=None,
                 book_id="", runtime=0, url="", img="", rss_url="",
                 copyright_year=0, language='english', from_data=None):
        """

        Args:
            title:
            authors:
            description:
            genres:
            book_id:
            runtime:
            url:
            rss_url:
            copyright_year:
            language:
            from_data:
        """
        AudioBook.__init__(self, title, authors, description, genres,
                           book_id, runtime, url, img, language)
        self.rss_url = rss_url
        self.copyright_year = copyright_year
        if from_data:
            self.from_json(from_data)
        self.raw = from_data or {}

    @property
    def description(self):
        """

        Returns:

        """
        return self._description.replace("<p>", "").replace("</p>", "") \
            .replace("(summary from Wikipedia)", "").strip().rstrip("\"") \
            .lstrip("\"")

    @property
    def rss_data(self):
        """

        Returns:

        """
        return feedparser.parse(self.rss_url)

    @property
    def streamer(self):
        """

        """
        for stream in self.rss_data["entries"]:
            try:
                yield stream['media_content'][0]["url"]
            except Exception as e:
                print(e)
                continue

    @property
    def authors(self):
        """

        Returns:

        """
        return [BookAuthor(from_data=a) for a in self._authors]

    @property
    def genres(self):
        """

        Returns:

        """
        return [BookGenre(from_data=a) for a in self._genres]

    def from_json(self, json_data):
        """

        Args:
            json_data:
        """
        AudioBook.from_json(self, json_data)
        self.url = json_data.get("url_librivox", self.url)
        self.runtime = json_data.get("totaltimesecs", self.runtime)
        self.copyright_year = json_data.get("copyright_year",
                                            self.copyright_year)
        self.rss_url = json_data.get("url_rss", self.rss_url)

    def __repr__(self):
        """

        Returns:

        """
        return "LibrivoxAudioBook(" + str(self) + ", " + self.book_id + ")"


class Librivox(AudioBookSource):
    """
    """
    base_url = "https://librivox.org/api/feed/audiobooks/?%s&format=json"
    authors_url = "https://librivox.org/api/feed/authors/?%s&format=json"

    @staticmethod
    def scrap_all_audiobooks(limit=2000, offset=0):
        """

        Generator, yields LibrivoxAudioBook objects

        Args:
            limit:
            offset:
        """
        url = Librivox.base_url % \
              ("limit=" + str(limit) + "offset=" + str(offset) + "&extended=1")
        json_data = requests.get(url).json()['books']
        for k in json_data:
            yield LibrivoxAudioBook(from_data=json_data[k])

    @staticmethod
    def get_audiobook(book_id):
        """

        Args:
            book_id:

        Returns:
            LibrivoxAudioBook

        """
        url = Librivox.base_url % ("id=" + str(book_id),)
        json_data = requests.get(url).json()['books']
        return LibrivoxAudioBook(from_data=json_data[0])

    @staticmethod
    def get_author(author_id):
        """

        Args:
            author_id:

        Returns:

        """
        url = Librivox.authors_url % ("id=" + str(author_id),)
        json_data = requests.get(url).json()["authors"]
        return BookAuthor(from_data=json_data[0])

    @staticmethod
    def search_audiobooks(since=None, author=None, title=None, genre=None,
                          limit=25):
        """

        Args:
            since: a UNIX timestamp; returns all projects cataloged since that time
            author: all records by that author last name
            title: all matching titles
            genre: all projects of the matching genre

        Returns:
            list : list of LibrivoxAudioBook objects
        """
        searchterm = []
        if limit:
            # TODO validate
            searchterm.append("limit=" + str(limit))
        if since:
            # TODO validate
            searchterm.append("since=" + since)
        if author:
            searchterm.append("author=" + author)
        if title:
            searchterm.append("title=" + title)
        if genre:
            # TODO validate
            searchterm.append("genre=" + genre)
        if not searchterm:
            raise TypeError
        searchterm = "&".join(searchterm)
        url = Librivox.base_url % (searchterm,)
        json_data = requests.get(url).json()["books"]
        return [LibrivoxAudioBook(from_data=a) for a in json_data]


if __name__ == "__main__":
    book = Librivox.search_audiobooks(title="War of the worlds")[0]
    book.play_mplayer()



import requests
import math
from bs4 import BeautifulSoup
from itertools import cycle
from audiobooker.utils.proxies import get_proxies, random_user_agent, \
    USER_AGENTS
from audiobooker.exceptions import RateLimitedError


class GoogleSearch(object):
    SEARCH_URL = "https://google.com/search"
    TOTAL_SELECTOR = "#resultStats"
    RESULTS_PER_PAGE = 10
    RATE_LIMITED = "Our systems have detected unusual traffic from your computer"
    PROXIES = []
    PROXY_POOL = cycle(PROXIES)
    USER_AGENT_POOL = cycle(USER_AGENTS)
    CURRENT_PROXY = None
    CURRENT_USER_AGENT = None
    PROXY_FALLBACK = False

    def __init__(self, proxied=False, proxy_list=None, fallback=False):

        if proxied:
            self.PROXY_FALLBACK = True
            self.next_proxy()
        elif fallback:
            self.PROXY_FALLBACK = True

        if proxy_list:
            GoogleSearch.PROXIES = proxy_list or []

    @staticmethod
    def next_proxy():
        # Get a proxy from the pool
        if not len(GoogleSearch.PROXIES):
            GoogleSearch.reset_proxies()
        GoogleSearch.CURRENT_PROXY = next(GoogleSearch.PROXY_POOL)
        GoogleSearch.CURRENT_USER_AGENT = next(GoogleSearch.USER_AGENT_POOL)

    @staticmethod
    def reset_proxies():
        print("refreshing proxy list")
        GoogleSearch.PROXIES = get_proxies(25)
        GoogleSearch.PROXY_POOL = cycle(GoogleSearch.PROXIES)

    @staticmethod
    def remove_proxy(proxy):
        if proxy in GoogleSearch.PROXIES:
            print("removing bad proxy", proxy)
            GoogleSearch.PROXIES.remove(proxy)
            GoogleSearch.PROXY_POOL = cycle(GoogleSearch.PROXIES)
            if GoogleSearch.CURRENT_PROXY == proxy:
                GoogleSearch.next_proxy()

    @staticmethod
    def make_request(url, data=None, user_agent=None, proxy=None):
        try:
            user_agent = user_agent or GoogleSearch.CURRENT_USER_AGENT
            if user_agent is None:
                user_agent = random_user_agent()
            proxy = proxy or GoogleSearch.CURRENT_PROXY
            if proxy:
                print("using proxy", proxy)
            if data is not None and proxy is not None:
                response = requests.get(url, data,
                                        headers={
                                            'User-Agent': user_agent,
                                            "Accept-Language": "en-US,en;q=0.5"
                                        },
                                        proxies={"http": proxy,
                                                 "https": proxy})
            elif data is not None:
                response = requests.get(url, data,
                                        headers={
                                            'User-Agent': user_agent,
                                            "Accept-Language": "en-US,en;q=0.5"
                                        })
            elif proxy is not None:
                response = requests.get(url,
                                        headers={
                                            'User-Agent': user_agent,
                                            "Accept-Language": "en-US,en;q=0.5"
                                        },
                                        proxies={"http": proxy,
                                                 "https": proxy})
            else:
                response = requests.get(url,
                                        headers={
                                            'User-Agent': user_agent,
                                            "Accept-Language": "en-US,en;q=0.5"
                                        })
        except Exception as e:
            print(e)
            GoogleSearch.remove_proxy(proxy)
            raise
        return response.text

    @staticmethod
    def get_soup(url, data=None):
        try:
            html = GoogleSearch.make_request(url, data)
            soup = BeautifulSoup(html, "html.parser")
            if GoogleSearch.RATE_LIMITED in soup.text:
                raise RateLimitedError("Google banned IP address")
            return soup
        except RateLimitedError as e:
            if GoogleSearch.PROXY_FALLBACK:
                GoogleSearch.remove_proxy(GoogleSearch.CURRENT_PROXY)
                if len(GoogleSearch.PROXIES):
                    return GoogleSearch.get_soup(url, data)
                raise RateLimitedError("Google banned IP of all proxies")
            raise

    @staticmethod
    def total_results(query):
        soup = GoogleSearch.get_soup(GoogleSearch.SEARCH_URL, {"q": query})
        try:
            totalText = \
                soup.select(GoogleSearch.TOTAL_SELECTOR)[0].text.split("(")[0]
        except IndexError:
            GoogleSearch.next_proxy()
            if len(GoogleSearch.PROXIES):
                return GoogleSearch.total_results(query)
            raise
        total = int("".join([c for c in totalText if c.isdigit()]))
        return total, soup

    @staticmethod
    def search(query, num_results=-1):
        total, soup = GoogleSearch.total_results(query)
        if num_results < 1:
            pages = int(
                math.ceil(total // float(GoogleSearch.RESULTS_PER_PAGE)))
        else:
            pages = int(
                math.ceil(num_results // float(GoogleSearch.RESULTS_PER_PAGE)))
        if pages < 1:
            pages = 1
        result_num = 0
        for i in range(pages):
            start = i * GoogleSearch.RESULTS_PER_PAGE

            params = {"q": query}
            if start > 0:
                params["start"] = start
                soup = GoogleSearch.get_soup(GoogleSearch.SEARCH_URL, params)
            results = GoogleSearch.parse_results(soup)
            result_num += len(results)

            if result_num > num_results and num_results > -1:
                d = result_num - num_results
                results = results[:len(results) - d]

            for idx, r in enumerate(results):
                r.num = idx + start
                r.total = total
                yield r

            if result_num > num_results and num_results > -1:
                return

    @staticmethod
    def parse_results(soup):
        bucket = []
        for a in soup.find_all("a"):
            try:
                url = a["href"]
                if url.startswith("/url?q="):
                    url = url.replace("/url?q=", "")
                elif url.startswith("/search?ie=UTF-8&q=related:"):
                    url = url.replace("/search?ie=UTF-8&q=related:", "")
                if not url.startswith("http"):
                    continue
                if "google" in url:
                    continue
                title = a.text.replace(url, "")
                bucket.append(SearchResult(title, url))
            except:
                pass

        return bucket


class SearchResult(object):
    def __init__(self, title, url, num=1, total=1):
        self.title = title
        self.url = url
        self.num = num
        self.total = 1

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return self.__str__()


if __name__ == "__main__":
    query = "lovecraft audio drama"
    count = 5
    google = GoogleSearch()
    print("Fetching first " + str(count) + " results for \"" + query + "\"...")
    for result in google.search(query, count):
        print("result #", result.num, " of ", result.total)
        print(result.url)
        print(result.title)



class AudioBookerException(Exception):
    pass


class UnknownBookIdException(AudioBookerException):
    pass


class UnknownGenreIdException(AudioBookerException):
    pass


class UnknownAuthorIdException(AudioBookerException):
    pass


class UnknownBookException(AudioBookerException):
    pass


class UnknownGenreException(AudioBookerException):
    pass


class UnknownAuthorException(AudioBookerException):
    pass


class ParseErrorException(AudioBookerException):
    pass


class UnknownDurationError(AudioBookerException):
    pass


class ScrappingError(AudioBookerException):
    pass


class RateLimitedError(ScrappingError, ConnectionRefusedError):
    """Google banned IP address"""



from audiobooker.exceptions import UnknownAuthorIdException, \
    UnknownBookIdException, UnknownAuthorException, UnknownBookException, \
    ParseErrorException

from audiobooker import AudioBook
from audiobooker.scrappers import AudioBookSource

from rapidfuzz import process


class KiddieAudioBook(AudioBook):

    def __init__(self, title="", authors=None, description="", genres=None,
                 book_id="", runtime=0, stream=None, url="", img="",
                 release_date=0, language='english', from_data=None):
        self.stream = stream
        self.release_date = release_date
        AudioBook.__init__(self, title, authors, description, genres,
                           book_id, runtime, url, img, language, from_data)

    @property
    def streamer(self):
        """

        """
        if not self.stream:
            raise ParseErrorException
        if isinstance(self.stream, list):
            for s in self.stream:
                yield s
            return
        yield self.stream

    @property
    def as_json(self):
        bucket = self.raw
        bucket["url"] = self.url
        bucket["img"] = self.img
        bucket["title"] = self.title
        bucket["authors"] = self._authors
        bucket["description"] = self._description
        bucket["format"] = self._description
        bucket["genres"] = self._genres
        bucket["id"] = self.book_id
        bucket["runtime"] = self.runtime
        bucket["language"] = self.lang
        bucket["release_date"] = self.release_date
        bucket["streams"] = [s for s in self.streamer]
        return bucket

    def calc_runtime(self, data=None):
        data = data or self.raw
        try:
            runtime = data["duration"].split(":")
            if len(runtime) == 1:  # seconds
                self.runtime += int(runtime[0])
            elif len(runtime) == 2:  # minutes : seconds
                self.runtime += int(runtime[1]) + (int(runtime[0]) * 60)
            elif len(runtime) == 3:  # hours : minutes : seconds
                self.runtime += int(runtime[2]) + (int(runtime[1]) * 60) + \
                                (int(runtime[0]) * 120)
        except:
            pass

    def from_json(self, json_data):
        """

        Args:
            json_data:
        """
        AudioBook.from_json(self, json_data)
        if not self.title:
            self.title = " ".join(json_data.get("titles", []))
        self._description = json_data.get("description", json_data.get(
            "format", self._description))
        self.book_id = json_data.get("release_id", self.book_id)
        self.stream = json_data.get("stream", self.stream)
        self.release_date = json_data.get("release_date", self.release_date)
        self.calc_runtime(json_data)

    def __repr__(self):
        """

        Returns:

        """
        return "KiddieRecordsAudioBook(" + self.__str__() + ", " + \
               self.book_id + ")"


class KiddieRecords(AudioBookSource):
    """
        Warning: this website formatting is kinda ridiculous, if some stuff
        bellow seems weird it's because it is

    """
    base_url = "http://www.kiddierecords.com/"
    urls = ["http://www.kiddierecords.com/2005/index.htm",
            "http://www.kiddierecords.com/2006/index.htm",
            "http://www.kiddierecords.com/2007/index.htm",
            "http://www.kiddierecords.com/2009/index.htm",
            "http://www.kiddierecords.com/mgac/index_1.htm",
            "http://www.kiddierecords.com/mgac/index_2.htm",
            "http://www.kiddierecords.com/albums/index_1.htm",
            "http://www.kiddierecords.com/albums/index_2.htm",
            "http://www.kiddierecords.com/albums/index_3.htm",
            "http://www.kiddierecords.com/xmas/index.htm"]

    @staticmethod
    def scrap_all_audiobooks(limit=-1, offset=0):
        """

        Generator, yields AudioBook objects


        Args:
            limit:
            offset:
        """

        for url in KiddieRecords.urls:
            buckets = []

            soup = KiddieRecords._get_soup(KiddieRecords._get_html(url))

            rows = soup.find_all("td")[2:]

            for idx, row in enumerate(rows):
                if not row:
                    continue
                text = row.find("p")
                if not text:
                    continue
                title = text.find_all("b")
                if not title:
                    continue

                # beware, there be dragons, data is not saved in a
                # structured way, website design is ridiculous, you can
                # probably waste time coming up with better rules
                bucket = {"url": url}
                published = author = release_format = duration = ""

                if len(title) > 2:
                    title = title[:2]

                # handle random \n that make no sense
                title = "\n".join([t.text for t in title])

                # handle \n that seem to be 2 titles in one release, maybe
                bucket["titles"] = [t.strip() for t in title.split("\n")]
                text = text.text.replace(title, "").strip()

                # regular file format? nah, at some point they gave up and
                # started adding youtube videos instead of real streams
                if 'Launch Stream' in text or 'Youtube' in text:
                    # some entries decided to include the streams inside the
                    # bold tag that usually is reserved for titles...
                    text = text.replace('Launch Stream', "") \
                        .replace('| Download Zip File', "")

                # randomly get empty strings because of stray \n ....
                components = [a.strip() for a in text.split("\n") if a.strip()]

                if len(components) == 0:
                    # these ones have their own div!
                    title = None
                    for p in row.find_all("font", {"size": 1}):
                        p = " ".join([a.strip() for a in p.text.split("\n")])
                        if not title:
                            title = p
                            bucket["titles"] = [t.strip() for t in
                                                title.split("\n")]
                        else:
                            author = p
                    # TODO scrap other fields not in soup
                    # failing mainly in http://www.kiddierecords.com/mgac/index_2.htm

                elif len(components) == 4:
                    author, published, release_format, duration = components
                # at some point putting all info also became optional
                elif len(components) == 3:
                    author, published, duration = components
                else:

                    # yep, even less info in a couple of them, can't depend
                    # on order...
                    for idx, c in enumerate(components):
                        c = c.strip()
                        if c.startswith("(1)") or c.startswith("(2)"):
                            release_format = c
                            components[idx] = ""
                        elif "Total Time:" in c:
                            duration = c
                            components[idx] = ""

                    # let's ensure we didn't miss other misbehaving entries
                    components = [c for c in components if c]
                    if len(components):
                        raise ParseErrorException

                bucket["authors"] = author.strip().split(" and ")
                if "©" in published:
                    bucket["release_id"], bucket["release_date"] = \
                        published.strip().split("©")
                else:
                    bucket["release_id"] = published.strip()
                bucket["format"] = release_format.strip()
                bucket["duration"] = duration \
                    .replace("Total Time: ", "").strip()

                # save for building AudioBook object later
                if bucket not in buckets:
                    buckets.append(bucket)

            # lazy find all streams by file extension,
            # assuming 1-to-1 match with buckets
            # NOTE bad assumption, corner cases creeped up
            el = soup.find_all("a")[8:]

            streams = [a["href"] for a in el if
                       ".zip" not in a["href"]
                       and not a["href"].endswith(".htm")]

            # lazy find all pics
            #  1-to-1 match with buckets
            el = soup.find_all("img")[1:]
            _index = url.split("/")[-1]
            pics = [url.replace(_index, pic["src"]) for pic in el]

            if not len(streams) == len(pics) == len(buckets):
                # NOTE Hack mentioned above, unlike with pics we can
                # not assure 1-to-1 match between buckets and streams
                # TODO FIX ME with proper parsing
                streams, pics, buckets = KiddieRecords._hack_fix_parsing(
                    streams, pics, buckets, url)

            # re check if fixes fixed everything
            if len(streams) == len(pics) == len(buckets):
                # everything checks out
                for idx, bucket in enumerate(buckets):
                    bucket["stream"] = streams[idx]
                    bucket["img"] = pics[idx]

                    yield KiddieAudioBook(from_data=bucket)
            else:
                raise ParseErrorException

    @staticmethod
    def _hack_fix_parsing(streams, pics, buckets, url):
        """ TODO FIX ME , bad asumption of 1-to-1 correspondence,
        that was true for most part/pages but not always"""

        if len(buckets) > len(pics):
            # parsed too many buckets
            raise ParseErrorException

        if len(streams) < len(buckets):
            # one entry is grayed out and has no stream
            if url == "http://www.kiddierecords.com/2007/index.htm":
                streams.insert(21, " ")
            else:
                raise ParseErrorException
        elif len(streams) > len(buckets):

            # Uncle Dave's Kiddie Christmas Collections entry messes
            # things up, it has 3 streams, we need to merge them into a
            # single index
            volumes = [a for a in streams if "/streams/volume_" in a]
            if len(volumes):
                streams.remove(volumes[1])
                streams.remove(volumes[2])
                streams[streams.index(volumes[0])] = volumes

                # in this same page another entry breaks things by randomly
                # deciding to name the stream differently, Spike Jones |
                # Jimmy Boyd, we need to merge these 2 streams
                bad_streans = ["http://youtu.be/oTLPMdsGgeQ",
                               "http://youtu.be/q8E1WafpFvo"]
                streams.remove(bad_streans[0])
                streams[streams.index(bad_streans[1])] = bad_streans
                # print(streams.index(bad_streans))  # should be 5

            if url == "http://www.kiddierecords.com/mgac/index_1.htm":
                # youtube + streams in some entries
                groups = [
                    [
                        "http://www.kiddierecords.com/mgac/streams/Dennis_the_Menace.m3u",
                        "http://youtu.be/Vlj7vqSDbqo"],
                    [
                        "http://www.kiddierecords.com/mgac/streams/The_Brave_Engineer.m3u",
                        "http://www.youtube.com/watch?v=SIrjnWKM9Tw"],
                    [
                        "http://www.kiddierecords.com/mgac/streams/Ferdinand_the_Bull.m3u",
                        "http://www.youtube.com/watch?v=0nSBxrFsLTY"],
                    [
                        "http://www.kiddierecords.com/mgac/streams/Carbon_the_Copy_Cat.m3u",
                        "http://youtu.be/_hg2ocEMET4"]
                ]
                for group in groups:
                    streams.remove(group[1])
                    streams[streams.index(group[0])] = group

            elif url == "http://www.kiddierecords.com/mgac/index_2.htm":

                # TODO component parsing for this page needs improvement
                # since its failing on a few fields it is considering a
                # duplicate entry for different Snow White and the Seven
                # Dwarfs releases
                missing_bucket = {
                    'authors': ['Lyn Murray Orchestra and Chorus'],
                    'duration': '23:25',
                    'format': '(4) 10" 78RPM record album with booklet',
                    'release_id': 'Decca A-368',
                    'titles': ['Snow White and the Seven Dwarfs'],
                    'url': 'http://www.kiddierecords.com/mgac/index_2.htm'}
                buckets.insert(69, missing_bucket)

        return streams, pics, buckets

    @staticmethod
    def get_audiobook(book_id):
        """

        Args:
            book_id:

        Returns:
            AudioBook

        """
        for b in KiddieRecords.get_all_audiobooks(cache=True):
            if b.book_id == book_id:
                return b
        raise UnknownBookIdException

    @staticmethod
    def get_author(author_id):
        """

        Args:
            author_id:

        Returns:
            BookAuthor
        """
        raise UnknownAuthorIdException

    @staticmethod
    def get_audiobook_id(book):
        """

        Args:
            book:

        Returns:

            book_id (str)
        """
        for b in KiddieRecords.get_all_audiobooks(cache=True):
            if b.title == book:
                return b
        raise UnknownBookException

    @staticmethod
    def get_author_id(author):
        """

        Args:
            author_id:

        Returns:
            author id (str)
        """
        raise UnknownAuthorException

    def search_audiobooks(self, since=None, author=None, title=None,
                          genre=None, limit=25):
        """

        Args:
            since: a year; returns all projects released since that time
            author: all records by that author last name
            title: all matching titles
            genre: all projects of the matching genre

        Yields:
            AudioBook objects
        """
        # priority for title matches
        alll = self.get_all_audiobooks()
        if title:
            for res in process.extract(title, alll, limit=limit):
                match, score = res
                yield match
                alll.remove(match)

        # second author matches
        if author:
            choices = [" ".join([str(a) for a in b.authors]) for b in alll]
            for res in process.extract(author, choices, limit=limit):
                match, score = res
                match = alll[choices.index(match)]
                yield match
                alll.remove(match)

        # TODO genre, somehow need to catalog songs vs stories etc.
        # comparing same as title for now
        if genre:
            for res in process.extract(genre, alll, limit=limit):
                match, score = res
                yield match
                alll.remove(match)


if __name__ == "__main__":
    from pprint import pprint

    scraper = KiddieRecords()
    for b in scraper.search_audiobooks(
            title="Snow White and the Seven Dwarfs"):
        print(b.as_json)

    for b in KiddieRecords.scrap_all_audiobooks():
        pprint(b.as_json)



import random

USER_AGENTS = [
    ('Mozilla/5.0 (X11; Linux x86_64) '
     'AppleWebKit/537.36 (KHTML, like Gecko) '
     'Chrome/57.0.2987.110 '
     'Safari/537.36'),  # chrome
    ('Mozilla/5.0 (X11; Linux x86_64) '
     'AppleWebKit/537.36 (KHTML, like Gecko) '
     'Chrome/61.0.3163.79 '
     'Safari/537.36'),  # chrome
    ('Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:55.0) '
     'Gecko/20100101 '
     'Firefox/55.0'),  # firefox
    ('Mozilla/5.0 (X11; Linux x86_64) '
     'AppleWebKit/537.36 (KHTML, like Gecko) '
     'Chrome/61.0.3163.91 '
     'Safari/537.36'),  # chrome
    ('Mozilla/5.0 (X11; Linux x86_64) '
     'AppleWebKit/537.36 (KHTML, like Gecko) '
     'Chrome/62.0.3202.89 '
     'Safari/537.36'),  # chrome
    ('Mozilla/5.0 (X11; Linux x86_64) '
     'AppleWebKit/537.36 (KHTML, like Gecko) '
     'Chrome/63.0.3239.108 '
     'Safari/537.36'),  # chrome
    ("Mozilla/5.0 (Windows NT 6.1; WOW64) "
     "AppleWebKit/537.36 (KHTML, like Gecko) "
     "Chrome/ 58.0.3029.81 Safari/537.36"),
]


def random_user_agent():
    return random.choice(USER_AGENTS)


try:
    import asyncio
    from proxybroker import Broker


    def get_proxies(n=10, types=['HTTP', 'HTTPS']):
        proxy_list = []
        proxies = asyncio.Queue()
        broker = Broker(proxies)

        async def show(proxies):
            while True:
                proxy = await proxies.get()
                if proxy is None:
                    break
                print('Found proxy: %s' % proxy)
                proxy_list.append(proxy.host + ":" + str(proxy.port))

        tasks = asyncio.gather(
            broker.find(types=types, limit=n),
            show(proxies))

        loop = asyncio.get_event_loop()
        loop.run_until_complete(tasks)
        return proxy_list

except ImportError:
    import requests
    from bs4 import BeautifulSoup


    def get_proxies(n=10, types=['HTTP', 'HTTPS']):
        url = 'https://free-proxy-list.net/'
        response = requests.get(url)
        soup = BeautifulSoup(response.text, "html.parser")
        proxies = set()
        for i in soup.find_all("tr")[1:]:

            if not i.text or not i.text[0].isdigit():
                break
            ip, port = [proxy.text for proxy in i.find_all("td")[:2]]
            proxy = ":".join([ip, port])
            proxies.add(proxy)
        return proxies


    import sys

    if sys.version.startswith("2"):

        try:
            import grey_harvest  # py2 only


            def get_proxies(n=10, types=['HTTP', 'HTTPS']):
                ''' spawn a harvester '''
                harvester = grey_harvest.GreyHarvester()

                ''' harvest some proxies from teh interwebz '''
                count = 0
                proxies = []
                for proxy in harvester.run():
                    proxies.append(proxy)
                    count += 1
                    if count >= n:
                        break
                return proxies

        except ImportError:
            pass



import json
import subprocess
import requests
from bs4 import BeautifulSoup

from audiobooker.exceptions import UnknownAuthorIdException, \
    UnknownBookIdException, UnknownDurationError, ScrappingError, \
    UnknownGenreIdException, UnknownAuthorException, UnknownBookException, \
    UnknownGenreException, ParseErrorException


class BookGenre(object):
    """
    """

    def __init__(self, name="", genre_id="", url="", from_data=None):
        """

        Args:
            name:
            genre_id:
            from_data:
        """
        self.name = name
        self.genre_id = genre_id
        self.url = url
        if from_data:
            self.from_json(from_data)

    @property
    def as_json(self):
        return {"name": self.name, "id": self.genre_id, "url": self.url}

    def from_json(self, json_data):
        """

        Args:
            json_data:
        """
        if isinstance(json_data, str):
            json_data = json.loads(json_data)
        if isinstance(json_data, BookGenre):
            json_data = json_data.as_json
        if not isinstance(json_data, dict):
            raise TypeError
        self.name = json_data.get("name", self.name)
        self.genre_id = json_data.get("id", self.genre_id)
        self.url = json_data.get("url", self.url)

    def __str__(self):
        """

        Returns:

        """
        return self.name

    def __repr__(self):
        """

        Returns:

        """
        return "BookGenre(" + str(self) + ", " + self.genre_id + ")"


class BookAuthor(object):
    """
    """

    def __init__(self, first_name="", last_name="", author_id="", url="",
                 from_data=None):
        """

        Args:
            first_name:
            last_name:
            author_id:
            from_data:
        """
        self.first_name = first_name
        self.last_name = last_name
        self.first_name, self.last_name = self.normalize_name()
        self.author_id = author_id
        self.url = url
        if from_data:
            self.from_json(from_data)

    def normalize_name(self):
        author = " ".join([self.first_name, self.last_name])
        names = author.split(" ")
        last_name = " ".join(names[1:])
        first_name = names[0]
        return first_name, last_name

    def from_json(self, json_data):
        """

        Args:
            json_data:
        """
        if isinstance(json_data, str):
            try:
                json_data = json.loads(json_data)
            except:
                json_data = {"last_name": json_data}
        if isinstance(json_data, BookAuthor):
            json_data = json_data.as_json
        if not isinstance(json_data, dict):
            print(json_data, type(json_data))
            raise TypeError
        self.first_name = json_data.get("first_name", self.first_name)
        self.last_name = json_data.get("last_name", self.last_name)
        self.first_name, self.last_name = self.normalize_name()
        self.author_id = json_data.get("id", self.author_id)
        self.url = json_data.get("url", self.url)

    @property
    def as_json(self):
        return {"first_name": self.first_name, "last_name": self.last_name,
                "id": self.author_id, "url": self.url}

    def __str__(self):
        """

        Returns:

        """
        return (self.first_name + " " + self.last_name).strip()

    def __repr__(self):
        """

        Returns:

        """
        return "BookAuthor(" + str(self) + ", " + self.author_id + ")"


class AudioBook(object):

    def __init__(self, title="", authors=None, description="", genres=None,
                 book_id="", runtime=0, url="", img="", language='english',
                 from_data=None, stream_list=None):
        """

        Args:
            title:
            authors:
            description:
            genres:
            book_id:
            runtime:
            language:
            from_data:
        """

        self.img = img
        self.url = url
        self.title = title
        self._authors = authors or []
        self._description = description
        self._genres = genres or []
        self.book_id = book_id
        self.runtime = runtime
        self.lang = language.lower()
        self._stream_list = stream_list or []
        if from_data:
            self.from_json(from_data)
        self.raw = from_data or {}

    def calc_runtime(self, data=None):
        raise UnknownDurationError

    def parse_page(self):
        raise ParseErrorException

    def from_page(self):
        self.raw = self.parse_page()

    @property
    def html(self):
        try:
            return requests.get(self.url).text
        except Exception as e:
            try:
                return requests.get(self.url, verify=False).text
            except:
                return None

    @property
    def soup(self):
        return BeautifulSoup(self.html, "html.parser")

    @property
    def description(self):
        """

        Returns:

        """
        return self._description.strip()

    @property
    def streamer(self):
        """

        """
        if self._stream_list is not None:
            return [s for s in self._stream_list]
        return []

    @property
    def streams(self):
        """

        Returns:

        """
        return [s for s in self.streamer]

    def play_sox(self):
        """

        """
        self.play("play %1")

    def play_mplayer(self):
        """

        """
        self.play("mplayer %1")

    def play_vlc(self):
        """

        """
        self.play("cvlc %1 --play-and-exit")

    def play(self, cmd="cvlc %1 --play-and-exit"):
        """

        Args:
            cmd:
        """
        for stream_url in self.streamer:
            print("playing", stream_url)
            if isinstance(cmd, str):
                cmd = cmd.split(" ")
            if isinstance(cmd, list):
                play_cmd = cmd
                for idx, c in enumerate(cmd):
                    if c == "%1":
                        play_cmd[idx] = stream_url
                subprocess.call(" ".join(play_cmd), shell=True)
            else:
                raise TypeError

    @property
    def authors(self):
        """

        Returns:

        """
        authors = []
        for a in self._authors:
            if isinstance(a, str):
                try:
                    a = json.loads(a)
                except Exception as e:
                    a = {"last_name": a}
            if isinstance(a, dict):
                authors += [a]
        return [BookAuthor(from_data=a) for a in authors]

    @property
    def genres(self):
        """

        Returns:

        """
        return [BookGenre(from_data=a) for a in self._genres]

    @property
    def as_json(self):
        bucket = self.raw
        bucket["url"] = self.url
        bucket["img"] = self.img
        bucket["title"] = self.title
        bucket["authors"] = self._authors
        bucket["description"] = self._description
        bucket["genres"] = self._genres
        bucket["id"] = self.book_id
        bucket["runtime"] = self.runtime
        bucket["language"] = self.lang
        bucket["streams"] = [s for s in self.streamer]
        return bucket

    def from_json(self, json_data):
        """

        Args:
            json_data:
        """
        if isinstance(json_data, str):
            json_data = json.loads(json_data)
        if not isinstance(json_data, dict):
            raise TypeError
        json_data = json_data or {}
        self.url = json_data.get("url", self.url)
        self.img = json_data.get("img",
                                 json_data.get("pic",
                                               json_data.get("image",
                                                             self.img)))
        self.title = json_data.get("title", json_data.get("name", self.title))
        self._authors = json_data.get("authors", self._authors)
        self._authors = self._authors or [json_data.get("author", "")]
        self._description = json_data.get("description", self._description)
        self._genres = json_data.get("genres", self._genres)
        self.book_id = json_data.get("id", self.book_id)
        self.runtime = json_data.get("runtime", self.runtime)
        self.lang = json_data.get('language',
                                  json_data.get('lang', self.lang)).lower()
        self._stream_list = json_data.get("streams", self._stream_list)
        self.raw = json_data

    def __str__(self):
        """

        Returns:

        """
        return self.title

    def __repr__(self):
        """

        Returns:

        """
        return "AudioBook(" + str(self) + ", " + self.book_id + ")"



import requests
from bs4 import BeautifulSoup
from threading import Thread
from rapidfuzz import process
from audiobooker.exceptions import UnknownAuthorIdException, \
    UnknownBookIdException, ScrappingError, UnknownGenreIdException, \
    UnknownAuthorException, UnknownBookException, UnknownGenreException
from audiobooker import AudioBook, BookAuthor
from audiobooker.utils.proxies import random_user_agent


class AudioBookSource(object):
    base_url = ""
    popular_url = ""
    genres_url = ""
    authors_url = ""
    search_url = ""
    _cache = None

    def populate_cache(self, books=None, threaded=False):
        if self._cache is None:
            if books:
                self._cache = books
                return
            if threaded:
                t = Thread(target=self.get_all_audiobooks,
                           daemon=True).start()
            else:
                self._cache = self.get_all_audiobooks()
        elif books:
            self._cache += books

    @property
    def genres(self):
        return sorted(['Advice', 'Instruction', 'Ancient Texts',
                       'Biography', 'Memoirs', 'Languages',
                       'Myths/Legends', 'Holiday', 'Art',
                       'Politics', 'Short stories', 'Romance',
                       'Essay/Short nonfiction', 'Fiction',
                       'Epistolary fiction', 'Science',
                       'Nature', 'Dramatic Works',
                       'Spy stories', 'History', 'Non-fiction',
                       'Historical Fiction', 'Play', 'Children',
                       'Satire', 'Humor',
                       'Classics (antiquity)', 'Travel',
                       'Religion', 'Adventure', 'Animals',
                       'Psychology', 'Sea stories',
                       'Horror/Ghost stories', 'Fantasy',
                       'Cookery', 'Poetry', 'Self Published',
                       'Westerns', 'Comedy', 'Music',
                       'Economics', 'Fairy tales', 'Tragedy',
                       'Teen/Young adult', 'Literature',
                       'War stories', 'Science fiction',
                       'Philosophy', 'Mystery'])

    @staticmethod
    def _get_html(url):
        user_agent = random_user_agent()
        try:
            return requests.get(url, headers={'User-Agent': user_agent}).text
        except Exception as e:
            return requests.get(url, verify=False,
                                headers={'User-Agent': user_agent}).text

    @staticmethod
    def _get_soup(html):
        return BeautifulSoup(html, "html.parser")

    @staticmethod
    def scrap_popular(limit=-1, offset=0):
        """

        Generator, yields AudioBook objects

        Args:
            limit:
            offset:
        """
        raise ScrappingError

    @staticmethod
    def scrap_all_audiobooks(limit=-1, offset=0):
        """

        Generator, yields AudioBook objects

        Args:
            limit:
            offset:
        """
        raise ScrappingError

    @staticmethod
    def scrap_by_genre(genre, limit=-1, offset=0):
        """

        Generator, yields AudioBook objects

        Args:
            genre:
            limit:
            offset:
        """
        raise ScrappingError

    def get_all_audiobooks(self, limit=2000, offset=0):
        """

        Args:
            limit:
            offset:

        Returns:
            list : list of LibrivoxAudioBook objects

        """

        if self._cache is not None:
            return self._cache
        self._cache = [book for book in self.scrap_all_audiobooks(limit,
                                                                  offset)]
        return self._cache

    @staticmethod
    def get_genre_id(genre):
        """

        Args:
            genre:

        Returns:
            genre_id (str)

        """
        raise UnknownGenreException

    @staticmethod
    def get_genre(genre_id):
        """

        Args:
            genre_id:

        Returns:
            BookGenre

        """
        raise UnknownGenreIdException

    @staticmethod
    def get_audiobook(book_id):
        """

        Args:
            book_id:

        Returns:
            AudioBook

        """
        raise UnknownBookIdException

    @staticmethod
    def get_author(author_id):
        """

        Args:
            author_id:

        Returns:
            BookAuthor
        """
        raise UnknownAuthorIdException

    @staticmethod
    def get_audiobook_id(book):
        """

        Args:
            book:

        Returns:

            book_id (str)
        """
        raise UnknownBookException

    @staticmethod
    def get_author_id(author):
        """

        Args:
            author_id:

        Returns:
            author id (str)
        """
        raise UnknownAuthorException

    def search_audiobooks(self, since=None, author=None, title=None,
                          genre=None, limit=25):
        """

        Args:
            since: a UNIX timestamp; returns all projects cataloged since that time
            author: all records by that author last name
            title: all matching titles
            genre: all projects of the matching genre
            limit: max entries to return (int)

        Returns:
            list : list of AudioBook objects
        """
        # priority for title matches
        alll = self.get_all_audiobooks()
        if title:
            for res in process.extract(title, alll, limit=limit):
                match, score = res
                yield match
                alll.remove(match)

        # second author matches
        if author:
            choices = [" ".join([str(a) for a in b.authors]) for b in alll]
            for res in process.extract(author, choices, limit=limit):
                match, score = res
                match = alll[choices.index(match)]
                yield match
                alll.remove(match)


if __name__ == "__main__":
    # read from csv or something
    streams = ['Aldous Huxley, Brave New World, '
               'https://1fizorq.oloadcdn.net/dl/l/lhcTuuSF1qQv_hV0/Q3JE4LxtblQ/16+-+Brave+New+World+-+Aldous+Huxley+-+1932.mp3?mime=true',
               'Arthur C. Clarke, Rendezvous with Rama, '
               'https://1fizors.oloadcdn.net/dl/l/-3TKoZ6X1GdzknSE/bPk3Nk2bvCs/14+-+Rendezvous+With+Rama+-+Arthur+C+Clarke+-+1973.mp3?mime=true',
               'Philip K. Dick, Do Androids Dream of Electric Sheep, '
               'https://1fizoro.oloadcdn.net/dl/l/PvHcbH-InPYzh6Bq/Yhc2d-us-kA/12+-+Do+Androids+Dream+of+Electric+Sheep+-+Philip+K+Dick+-+1968.mp3?mime=true',
               'George Orwell, Animal Farm, https://www.youtube.com/watch?v=4Ln-Bfg6Wk0',
               'George Orwell, 1984, '
               'https://1fizorp.oloadcdn.net/dl/l/C76B03TG8T9wi-2z/vBhqbIHW5iM/Nineteen+Eighty+Four+-+George+Orwell.mp3?mime=true',
               'Arthur C. Clarke, 2001 A Space Odyssey, '
               'https://1fizorm.oloadcdn.net/dl/l/5YLhBL8cXOiOVqDh/-3w7Z_VvYCU/8+-+2001%3B+A+Space+Odyssey+-+Arthur+C+Clarke+-+1968.mp3?mime=true',
               'Arthur C. Clarke, Childhood’s End, '
               'https://1fizors.oloadcdn.net/dl/l/FsqNl_M7s_s54OeE/q2HkT0EYx8w/18+-+Childhood%27s+End+-+Arthur+C+Clarke+-+1954.mp3?mime=true',
               'Robert A. Heinlein, Starshio Troopers, '
               'https://1fizorn.oloadcdn.net/dl/l/W6xGJG_Ig2l_O7s8/XXegbOEAdz8/9+-+Starship+Troopers+-+Robert+A+Heinlein+-+1959.mp3?mime=true',
               'Robert A. Heinlein, The Moon is a Harsh Mistress, '
               'https://1fizorp.oloadcdn.net/dl/l/iHmjhmlTmw5RvaXnF4/8bwjgLgz0o0/19+-+The+Moon+is+a+Harsh+Mistress+-+Robert+A+Heinlein+-+1966.mp3?mime=true'
               ]

    from pprint import pprint

    book_lib = AudioBookSource()

    for book in streams:
        author, title, stream = book.split(",")

        audio_book = AudioBook(description="awesome book",
                               from_data={"title": title,
                                          "authors": [author],
                                          "streams": [stream]})

        book_lib.populate_cache([audio_book])

        author = BookAuthor(from_data=author)

    for book in book_lib.search_audiobooks(title="androids", limit=1):
        print(book.title, book.authors)

    for book in book_lib.search_audiobooks(author="Heinlein", limit=3):
        pprint(book.as_json)



from setuptools import setup

setup(
    name='audiobooker',
    version='0.2.7',
    packages=['audiobooker', 'audiobooker.scrappers', 'audiobooker.utils'],
    install_requires=["requests", "bs4", "feedparser", "rapidfuzz"],
    url='https://github.com/OpenJarbas/audiobooker',
    license='MIT',
    author='jarbasAI',
    author_email='jarbasai@mailfence.com',
    description='audio book scrapper'
)



from audiobooker.exceptions import UnknownAuthorIdException, \
    UnknownBookIdException, UnknownAuthorException, UnknownBookException
from audiobooker import AudioBook, BookAuthor
from audiobooker.scrappers import AudioBookSource


class HPPodcraft(AudioBookSource):
    base_url = "http://hppodcraft.com/"
    search_url = "http://hppodcraft.com/full-story-readings/"

    def __init__(self):
        raise RuntimeError("Webpage source changed, scrapper deprecated ")

    @staticmethod
    def scrap_all_audiobooks(limit=-1, offset=0):
        """

        Generator, yields AudioBook objects

        Args:
            limit:
            offset:
        """
        soup = HPPodcraft._get_soup(
            HPPodcraft._get_html(HPPodcraft.search_url))
        i = 0
        for url in soup.find_all("a"):
            href = url["href"]
            if href.endswith(".mp3"):
                title = url["title"].replace("HP Lovecraft Literary Podcast "
                                             "- ", "")
                if not href.startswith(HPPodcraft.base_url):
                    href = HPPodcraft.base_url + href
                yield AudioBook(title=title, book_id=str(i),
                                stream_list=[href], url=HPPodcraft.search_url,
                                authors=[{"first_name": "Howard P.",
                                          "last_name": "Lovecraft",
                                          "id": "0",
                                          "url": "http://hppodcraft.com/"}])
                i += 1

    def get_audiobook(self, book_id):
        """

        Args:
            book_id:

        Returns:
            AudioBook

        """
        for b in self.get_all_audiobooks():
            if str(b.book_id) == str(book_id):
                return b
        raise UnknownBookIdException

    @staticmethod
    def get_author(author_id=0):
        """

        Args:
            author_id:

        Returns:
            BookAuthor
        """
        if str(author_id) == "0":
            return BookAuthor(from_data={"first_name": "Howard P.",
                                         "last_name": "Lovecraft",
                                         "id": "0",
                                         "url": "http://hppodcraft.com/"})
        raise UnknownAuthorIdException

    def get_audiobook_id(self, book):
        """

        Args:
            book:

        Returns:

            book_id (str)
        """
        for b in self.get_all_audiobooks():
            if b.title == book:
                return b.book_id
        raise UnknownBookException

    @staticmethod
    def get_author_id(author):
        """

        Args:
            author_id:

        Returns:
            author id (str)
        """
        if "lovecraft" in author.lower():
            return "0"
        raise UnknownAuthorException


if __name__ == "__main__":

    from pprint import pprint

    book_lib = HPPodcraft()

    for book in book_lib.scrap_all_audiobooks():
        pprint(book.as_json)
        book.play()
